from .deploy import Deployment as Deployment
from .deploy import deploy as deploy
from .deploy import dump_deployment as dump_deployment
from .deploy import load_vars as load_vars
from .deploy import undump_deployment as undump_deployment
from .http_client import create_client as create_client
from .log import list_deployments as list_deployments
from .log import setup as setup


# import all resources so that the registration gets run.
from fbnconfig import access
from fbnconfig import compliance
from fbnconfig import configuration
from fbnconfig import customentity
from fbnconfig import datatype
from fbnconfig import drive
from fbnconfig import identity
from fbnconfig import lumi
from fbnconfig import property
from fbnconfig import recipe
from fbnconfig import scheduler
from fbnconfig import sequence
from fbnconfig import side_definition
from fbnconfig import transaction_type



from __future__ import annotations

import datetime as dt
import json
from enum import StrEnum
from hashlib import sha256
from typing import Annotated, Any, Dict, List, Optional, Self, Sequence, Union

import httpx
from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, model_validator

from fbnconfig.property import LifeTime, PropertyKey
from fbnconfig.transaction_type import MetricValue

from .resource_abc import CamelAlias, Ref, Resource, register_resource


def des_isodate(value: str | dt.datetime) -> dt.datetime:
    """Deserialize ISO 8601 date string to datetime object."""
    if isinstance(value, dt.datetime):
        return value.astimezone(tz=dt.timezone.utc)
    return dt.datetime.fromisoformat(value)


def ser_isodate(value: dt.datetime) -> str:
    """Serialize datetime object to ISO 8601 string."""
    return value.isoformat()


IsoDateTime = Annotated[
    dt.datetime | dt.date,
    BeforeValidator(des_isodate), PlainSerializer(ser_isodate)
]


class PropertyValue(CamelAlias, BaseModel):
    property_key: PropertyKey = Field(serialization_alias="propertyKey")
    label_value: Optional[str] = Field(default=None, serialization_alias="labelValue")
    metric_value: Optional[MetricValue] = Field(default=None, serialization_alias="metricValue")
    label_set_value: Optional[List[str]] = Field(default=None, serialization_alias="labelSetValue")
    effective_from: IsoDateTime | None = None
    effective_until: IsoDateTime | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info):
        if not isinstance(data, dict):
            return data
        style = info.context.get("style", "api") if info.context else "api"
        if style == "api" and data.get("value", None):
            return data | data["value"]
        return data

    @model_validator(mode="after")
    def validate_one_value_exists(self) -> Self:
        fields = ["label_value", "metric_value", "label_set_value"]
        s = [field for field in fields if getattr(self, field) is not None]
        if len(s) > 1:
            raise KeyError(f"Cannot set {' and '.join(s)}, only one of {' or '.join(fields)} can be set")

        return self


class SupportedDomain(StrEnum):
    Instrument = "Instrument"
    Person = "Person"
    LegalEntity = "LegalEntity"
    CustomEntity = "CustomEntity"


def serialize_properties_as_dict(values: List[PropertyValue], info):
    style = info.context.get("style", "api") if info.context else "api"
    if style == "dump":
        return values
    property_dict = {}
    for prop in values:
        value = {}
        if prop.label_value:
            value["labelValue"] = prop.label_value
        elif prop.metric_value:
            value["metricValue"] = prop.metric_value
        elif prop.label_set_value:
            value["labelSetValue"] = prop.label_set_value
        key = f"{prop.property_key.domain.value}/{prop.property_key.scope}/{prop.property_key.code}"
        property_dict[key] = {"key": key, "value": value}

        if prop.effective_from is not None:
            property_dict[key]["effectiveFrom"] = prop.effective_from

        if prop.effective_until is not None:
            property_dict[key]["effectiveUntil"] = prop.effective_until

    return property_dict


def des_properties_dict(data, info) -> Sequence[PropertyValue] | Dict | None:
    if not isinstance(data, dict) or data is None:
        return data
    style = info.context.get("style", "api") if info.context else "api"
    if style == "api" and isinstance(data, dict):
        data = [value | {"propertyKey": value["key"]} for value in data.values()]
    return data


PropertyValueDict = Annotated[
    Sequence[PropertyValue],
    BeforeValidator(des_properties_dict), PlainSerializer(serialize_properties_as_dict)
]


@register_resource()
class IdentifierDefinitionRef(BaseModel, Ref):
    """
    Reference to an identifier definition resource.

    Example
    ----------
    >>> from fbnconfig import identifier_definition
    >>> ref = identifier_definition.IdentifierDefinitionRef(
    >>>  id="identifier-def-ref",
    >>>  domain="Instrument",
    >>>  identifier_scope="id_scope",
    >>>  identifier_type="id_type")
    """

    id: str = Field(exclude=True)
    domain: SupportedDomain
    identifier_scope: str
    identifier_type: str

    def attach(self, client):
        """Attach to an existing identifier definition resource."""
        scope = self.identifier_scope
        id_type = self.identifier_type
        try:
            client.get(f"/api/api/identifierdefinitions/{self.domain}/{scope}/{id_type}")
        except httpx.HTTPStatusError as ex:
            error_message = f"Identifier Definition {self.domain}/{scope}/{id_type} does not exist"
            if ex.response.status_code == 404:
                raise RuntimeError(error_message)
            else:
                raise ex


@register_resource()
class IdentifierDefinitionResource(CamelAlias, BaseModel, Resource):
    """identifier definition resource"""

    id: str = Field(exclude=True)
    domain: SupportedDomain
    identifier_scope: str
    identifier_type: str
    life_time: LifeTime
    hierarchy_usage: str | None = None
    hierarchy_level: str | None = None
    display_name: str | None = None
    description: str | None = None
    properties: PropertyValueDict | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        # Handle id from context (for dump/undump)
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}

        return data

    def __get_content_hash__(self) -> str:
        dump = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        return sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()

    def read(self, client: httpx.Client, old_state) -> Dict[str, Any]:
        domain = old_state.domain
        scope = old_state.identifier_scope
        id_type = old_state.identifier_type
        response = client.get(f"/api/api/identifierdefinitions/{domain}/{scope}/{id_type}")

        result = response.json()
        # Remove unnecessary fields
        result.pop("href", None)

        return result

    def create(self, client: httpx.Client) -> Dict[str, Any]:
        body = self.model_dump(mode="json", exclude_none=True, by_alias=True)

        response = client.post("/api/api/identifierdefinitions", json=body)
        result = response.json()

        # Remove unnecessary fields
        result.pop("href", None)

        source_version = self.__get_content_hash__()
        remote_version = result["version"]["asAtVersionNumber"]

        return {
            "domain": self.domain,
            "identifier_scope": self.identifier_scope,
            "identifier_type": self.identifier_type,
            "source_version": source_version,
            "remote_version": remote_version
        }

    def update(self, client: httpx.Client, old_state) -> Union[None, Dict[str, Any]]:
        current_identity_tuple = (self.domain, self.identifier_scope, self.identifier_type)
        old_state_tuple = (old_state.domain, old_state.identifier_scope, old_state.identifier_type)

        if current_identity_tuple != old_state_tuple:
            self.delete(client, old_state)
            return self.create(client)

        source_hash = self.__get_content_hash__()
        remote = self.read(client, old_state)
        remote_hash = remote["version"]["asAtVersionNumber"]

        if remote_hash == old_state.remote_version and source_hash == old_state.source_version:
            return None

        #  Exclude fields not needed in put call
        body = self.model_dump(
            mode="json",
            exclude_none=True,
            by_alias=True,
            exclude={"domain", "identifier_scope", "identifier_type"}
        )

        scope = self.identifier_scope
        id_type = self.identifier_type

        response = client.put(
            f"/api/api/identifierdefinitions/{self.domain}/{scope}/{id_type}",
            json=body)
        result = response.json()

        # Remove unnecessary fields
        result.pop("href", None)

        return {
            "domain": self.domain,
            "identifier_scope": self.identifier_scope,
            "identifier_type": self.identifier_type,
            "source_version": source_hash,
            "remote_version": result["version"]["asAtVersionNumber"]
        }

    @staticmethod
    def delete(client: httpx.Client, old_state) -> None:

        domain = old_state.domain
        scope = old_state.identifier_scope
        id_type = old_state.identifier_type
        client.delete(f"/api/api/identifierdefinitions/{domain}/{scope}/{id_type}")

    def deps(self):
        deps = []
        if self.properties is None:
            return deps

        for perp_prop in self.properties:
            deps.append(perp_prop.property_key)

        return deps



from fbnconfig import Deployment, datatype, property

"""
An example configuration for data types.
The script configures the following entities:
- Data type

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01743/
"""


def configure(env):
    strategy_type = datatype.DataTypeResource(
        id="datatype-example",
        scope="sc1",
        code="cd1",
        type_value_range=datatype.TypeValueRange.CLOSED,
        display_name="Portfolio Strategy Test",
        description="A test datatype modified",
        value_type=datatype.ValueType.STRING,
        reference_data=datatype.ReferenceData(
            field_definitions=[
                datatype.FieldDefinition(key="description", is_required=True, is_unique=False),
                datatype.FieldDefinition(key="commission", is_required=True, is_unique=False),
            ],
            values=[
                datatype.FieldValue(
                    value="I", fields={"description": "Investment Portfolio", "commission": "0.01"}
                ),
                datatype.FieldValue(
                    value="H", fields={"description": "Hedging Portfolio", "commission": "0.05"}
                ),
                datatype.FieldValue(
                    value="C", fields={"description": "Client Portfolio", "commission": "0.10"}
                ),
            ],
        ),
    )

    strategy_prop = property.DefinitionResource(
        id="strategy-property",
        domain=property.Domain.Portfolio,
        scope="sc1",
        code="strategy",
        display_name="Example portfolio strategy",
        data_type_id=strategy_type,
        constraint_style=property.ConstraintStyle.Property,
        property_description="Example strategy datatype property",
        life_time=property.LifeTime.Perpetual,
    )

    priority_type = datatype.DataTypeResource(
        id="priority",
        scope="sc1",
        code="priority",
        type_value_range=datatype.TypeValueRange.CLOSED,
        display_name="Priority datatype example",
        description="A datatype for Priority values",
        value_type=datatype.ValueType.STRING,
        acceptable_values=["High", "Medium", "Low"],
    )

    return Deployment("datatype_example", [strategy_type, strategy_prop, priority_type])



import os
from types import SimpleNamespace

from pytest import fixture

import fbnconfig
from fbnconfig import compliance as cp
from fbnconfig import property as pr
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture()
def base_resources(setup_deployment):
    deployment_name = setup_deployment.name
    # Create property definition reference
    ccy = pr.DefinitionRef(
        id="prop_ccy",
        domain=pr.Domain.Instrument,
        scope="default",
        code="Currency",
    )
    # Create compliance template
    template1 = cp.ComplianceTemplateResource(
        id="template1",
        scope=deployment_name,
        code="template1",
        tags=["tag1", "tag2"],
        description="This is the first compliance template",
        variations=[
            cp.ComplianceTemplateVariation(
                label="Variation1",
                description="This is the first variation of the template",
                outcome_description="This variation checks for specific conditions",
                steps=[
                    cp.BranchStep(
                        label="Branch-Step-1",
                        parameters=[
                            cp.ComplianceTemplateParameter(
                                name="param1",
                                description="Parameter for branch step",
                                type="string",
                            )
                        ]
                    ),
                ],
                referenced_group_label=None
            )
        ]
    )
    # Create compliance rule
    rule1 = cp.ComplianceRuleResource(
        id="rule1",
        scope=deployment_name,
        code="rule1",
        name="ExampleComplianceRule",
        description="This is an example compliance rule",
        active=True,
        template_id=template1,
        variation="Variation1",
        portfolio_group_id=cp.ResourceId(
            scope="robtest",
            code="group10"
        ),
        parameters={
            "Branch-Step-1.BranchingKey": cp.GroupBySelectorComplianceParameter(value="1"),
        },
        properties=[
            cp.PropertyListItem(
                key=ccy,
                value=cp.PropertyValue(
                    label_value="USD",
                ),
            )
        ]
    )
    return [template1, rule1]


@fixture()
def setup_deployment():
    deployment_name = gen("compliance")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Clean up compliance resources after tests
    try:
        fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
        client.delete(f"/api/api/compliance/rules/{deployment_name}/rule1")
        client.delete(f"/api/api/compliance/templates/{deployment_name}/template1")
    except Exception:
        pass  # Ignore cleanup errors


def test_create(setup_deployment, base_resources):
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # Verify template was created
    template_response = client.get(
        f"/api/api/compliance/templates/{setup_deployment.name}/template1"
    ).json()
    assert template_response["id"]["scope"] == setup_deployment.name
    assert template_response["id"]["code"] == "template1"
    assert template_response["description"] == "This is the first compliance template"
    assert len(template_response["variations"]) == 1
    assert template_response["variations"][0]["label"] == "Variation1"
    # Verify rule was created
    rule_response = client.get(f"/api/api/compliance/rules/{setup_deployment.name}/rule1").json()
    assert rule_response["id"]["scope"] == setup_deployment.name
    assert rule_response["id"]["code"] == "rule1"
    assert rule_response["name"] == "ExampleComplianceRule"
    assert rule_response["description"] == "This is an example compliance rule"
    assert rule_response["active"] is True
    assert rule_response["templateId"]["scope"] == setup_deployment.name
    assert rule_response["templateId"]["code"] == "template1"
    assert rule_response["variation"] == "Variation1"


def test_nochange(setup_deployment, base_resources):
    # given we have deployed the base case
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # when we apply it again
    update = fbnconfig.deploy(deployment, lusid_env, token)
    # then there are no changes
    template_changes = [a.change for a in update if a.type == "ComplianceTemplateResource"]
    rule_changes = [a.change for a in update if a.type == "ComplianceRuleResource"]
    assert template_changes == ["nochange"]
    assert rule_changes == ["nochange"]


def test_teardown(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    # given we have deployed the base case
    deployment = fbnconfig.Deployment(deployment_name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # when we remove all the resources
    empty = fbnconfig.Deployment(deployment_name, [])
    update = fbnconfig.deploy(empty, lusid_env, token)
    # then resources are removed
    template_changes = [a.change for a in update if a.type == "ComplianceTemplateResource"]
    rule_changes = [a.change for a in update if a.type == "ComplianceRuleResource"]
    assert template_changes == ["remove"]
    assert rule_changes == ["remove"]


def test_update(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    # Given we have deployed the base case
    initial = fbnconfig.Deployment(deployment_name, base_resources)
    fbnconfig.deploy(initial, lusid_env, token)
    # when we update resources
    ccy = pr.DefinitionRef(
        id="prop_ccy",
        domain=pr.Domain.Instrument,
        scope="default",
        code="Currency",
    )
    # Update the template with different description
    updated_template = cp.ComplianceTemplateResource(
        id="template1",
        scope=deployment_name,
        code="template1",
        tags=["tag1", "tag2", "updated"],  # Added tag
        description="This is the updated first compliance template",  # Changed description
        variations=[
            cp.ComplianceTemplateVariation(
                label="Variation1",
                description="This is the updated first variation of the template",  # Changed
                outcome_description="This variation checks for updated specific conditions",  # Changed
                steps=[
                    cp.BranchStep(
                        label="Branch-Step-1",
                        parameters=[
                            cp.ComplianceTemplateParameter(
                                name="param1",
                                description="Updated parameter for branch step",  # Changed description
                                type="string",
                            )
                        ]
                    ),
                ],
                referenced_group_label=None
            )
        ]
    )
    # Update the rule with different description
    updated_rule = cp.ComplianceRuleResource(
        id="rule1",
        scope=deployment_name,
        code="rule1",
        name="UpdatedExampleComplianceRule",  # Changed name
        description="This is an updated example compliance rule",  # Changed description
        active=False,  # Changed active status
        template_id=updated_template,
        variation="Variation1",
        portfolio_group_id=cp.ResourceId(
            scope="robtest",
            code="group10"
        ),
        parameters={
            # Changed value
            "Branch-Step-1.BranchingKey": cp.GroupBySelectorComplianceParameter(value="2"),
        },
        properties=[
            cp.PropertyListItem(
                key=ccy,
                value=cp.PropertyValue(
                    label_value="EUR",  # Changed currency
                ),
            )
        ]
    )
    updated_resources = [updated_template, updated_rule]
    # and deploy it
    updated_deployment = fbnconfig.Deployment(deployment_name, updated_resources)
    update = fbnconfig.deploy(updated_deployment, lusid_env, token)
    # then we expect both resources to change
    template_changes = [a.change for a in update if a.type == "ComplianceTemplateResource"]
    rule_changes = [a.change for a in update if a.type == "ComplianceRuleResource"]
    assert template_changes == ["update"]
    assert rule_changes == ["update"]
    # and they have the new values
    updated_template_response = client.get(
        f"/api/api/compliance/templates/{deployment_name}/template1"
    ).json()
    assert updated_template_response["description"] == "This is the updated first compliance template"
    updated_rule_response = client.get(
        f"/api/api/compliance/rules/{deployment_name}/rule1"
    ).json()
    assert updated_rule_response["name"] == "UpdatedExampleComplianceRule"
    assert updated_rule_response["description"] == "This is an updated example compliance rule"
    assert updated_rule_response["active"] is False



import csv
import hashlib
import io
import json
import textwrap
from datetime import datetime
from types import SimpleNamespace
from typing import Any, Dict, List, Optional
from unittest.mock import patch

import httpx
import pytest
from respx import Route, Router, mock

from fbnconfig import lumi
from fbnconfig.backoff_handler import BackoffHandler
from fbnconfig.lumi import TaskStatus
from fbnconfig.property import DefinitionRef, Domain

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.raise_for_status()


@pytest.fixture(scope="module", autouse=True)
def dont_sleep():
    with patch("time.sleep", return_value=None) as _fixture:
        yield _fixture


def setup_sql_background_routes(
    respx_mock: Router,
    content: Optional[str] = None,
    execution_id: str = "abcdefg",
    task_status: TaskStatus = TaskStatus.RAN_TO_COMPLETION,
    status_code: Optional[List[int]] = None,
    result_data: Optional[Any] = None,
) -> dict[str, Route]:
    routes = {
        "put": respx_mock.put(
            "/honeycomb/api/SqlBackground",
            # # don't pass values as arguments if they're set to none
            **dict(
                filter(
                    lambda item: item[1] is not None,
                    {"content": content, "headers": [("Content-type", "text/plain")]}.items(),
                )
            ),
        ).mock(
            side_effect=[
                httpx.Response(
                    status_code.pop() if status_code else 202, json={"executionId": execution_id}
                )
            ]
        ),
        "status": respx_mock.get(f"/honeycomb/api/SqlBackground/{execution_id}").mock(
            side_effect=[
                httpx.Response(status_code.pop() if status_code else 200, json={"status": task_status})
            ]
        ),
        "result": respx_mock.get(f"/honeycomb/api/SqlBackground/{execution_id}/jsonproper").mock(
            side_effect=[httpx.Response(status_code.pop() if status_code else 200, json=result_data)]
        ),
    }

    return routes


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeViewRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})
    view_attach_content = textwrap.dedent("""\
                select r.Version from sys.registration as r
                where name = 'views.test.view1'
                order by r.Version asc
                limit 1
            """)

    def test_attach_when_exists(self, respx_mock, monkeypatch):
        # given a view in lumi
        setup_sql_background_routes(
            respx_mock, content=self.view_attach_content, result_data=["one row"]
        )

        client = self.client
        # when we attach
        sut = lumi.ViewRef(id="abc", provider="views.test.view1")
        sut.attach(client)
        assert sut._version == "one row"

    def test_attach_when_not_exists(self, respx_mock):
        setup_sql_background_routes(respx_mock, content=self.view_attach_content, result_data=[])
        # given no matching views exist yet

        client = self.client
        # when we attach a view ref
        sut = lumi.ViewRef(id="abc", provider="views.test.view1")
        # then it raises
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Failed to attach ref to views.test.view1" in str(ex.value)

    @mock(assert_all_called=False)
    def test_attach_when_http_error(self, respx_mock):
        # given the query fails
        routes: dict[str, Route] = setup_sql_background_routes(
            respx_mock, content=self.view_attach_content, status_code=[400]
        )

        client = self.client
        # when we attach a view ref
        sut = lumi.ViewRef(id="abc", provider="views.test.view1")
        # http error is raised
        with pytest.raises(httpx.HTTPStatusError):
            sut.attach(client)
        assert routes["put"].call_count == 1

    def test_dump(self):
        sut = lumi.ViewRef(
            id="lumi-example-ref",
            provider="Views.fbnconfig.existing_view"
        )

        result = sut.model_dump(context={"style": "dump"})

        # id field is globally excluded, so it won't appear in dumps
        assert "id" not in result
        assert result["provider"] == "Views.fbnconfig.existing_view"


# constructs the row from sys.file
def make_sys_file(sql: str, metadata: Dict, version: int) -> Dict:
    return {"Content": f"{sql}--- MetaData --- " + json.dumps(metadata), "Version": version}


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeViewResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @staticmethod
    def setup_class():
        # reduce time waiting for registration
        lumi.ViewResource.Registration.tries = 3
        lumi.ViewResource.Registration.wait_time = 0

    def test_create(self, respx_mock):
        exec_id = "abcdefg"

        expected = textwrap.dedent("""\
                   @x = use Sys.Admin.SetupView
                   --provider="Views.Unit.Test"
                   --description="a view"
                   ----
                   select * from foo;
                   enduse;
                   select * from @x;
               """)
        # setup_sql_background_routes(respx_mock, content=expected, result_data={})

        expected_registration = textwrap.dedent("""\
                    select Version from sys.registration where Name='Views.Unit.Test'
                    order by Version asc
                    limit 1
                """)

        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )
        # setup_sql_background_routes(
        #     respx_mock,
        #     content=expected_registration,
        #     execution_id="123",
        #     result_data=[{"Version": 1}])

        # given a view to create
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo;", description="a view"
        )
        # when we do
        state = sut.create(self.client)
        # the state is returned
        respx_mock.assert_all_called()
        assert put.calls[0].request.content.decode("utf8") == expected
        assert put.calls[1].request.content.decode("utf8") == expected_registration
        assert state == {"provider": "Views.Unit.Test"}

    def test_create_waits_for_registration(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                # Query creation + 3x registration
                httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})
            ]
            * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={})]
            +
            # and the registration check was performed a second time
            # because no version came back on the first check
            [httpx.Response(200, json=[])] * 2
            + [httpx.Response(200, json=[{"Version": "1"}])]
        )

        expected = textwrap.dedent("""\
                    select Version from sys.registration where Name='Views.Unit.Test'
                    order by Version asc
                    limit 1
                """)

        # given a view to create
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo;", description="a view"
        )
        # when we do
        state = sut.create(self.client)
        respx_mock.assert_all_called()
        assert put.calls[1].request.content.decode("utf8") == expected
        # the state is returned
        assert state == {"provider": "Views.Unit.Test"}

    def test_create_completes_if_registration_incomplete(self, respx_mock, capsys):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                # Query creation + 3x registration
                httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})
            ]
            * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={})] + [httpx.Response(200, json=[])] * 3
        )

        # given a view to create
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo;", description="a view"
        )
        # when we do
        state = sut.create(self.client)
        # the state is returned even though the registration did not
        # complete in time
        assert state == {"provider": "Views.Unit.Test"}
        # and the registration check was performed 3 times
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        respx_mock.assert_all_called()
        assert put.calls[1].request.content.decode("utf8") == expected

        # check warning is returned
        captured = capsys.readouterr()
        assert captured.err == (
            f"warning: no view registration after "
            f"{lumi.ViewResource.Registration.tries}"
            f" tries for Views.Unit.Test"
        )

    def test_create_with_saved_option(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired view with one of the saved options
        # (ones which impact the stored view in lumi)
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            documentation_link="http://example.com",
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # and a create request was sent with an =
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --documentationLink="http://example.com"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_with_saved_bool_option(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired state with a boolean saved option
        # variableShape
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            variable_shape=True,
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the create adds the option without an =
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --variableShape
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_with_test_option(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given desired includes limit, an option which is used for
        # running the view but does not get saved
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            limit=42,
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the test option gets sent and is not quoted because
        # it's an int
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --limit=42
            --description="a view"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_parameter(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a parameter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=23,
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the parameter is added at the end of the options
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,23,false,"p1 is nice"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_parameter_escaped(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a text parameter where the value
        # includes a double quote
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Text,
                    value='embedded " quote',
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the text is wrapped in quotes and the embedded quote
        # gets escaped
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Text,"embedded "" quote",false,"p1 is nice"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_parameter_no_tooltip(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a parameter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1", type=lumi.ParameterType.Int, value=23, set_as_default_value=False
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the parameter is added at the end of the options
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,23,false
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_scalar_variable(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a variable which is the default value
        # for a parameter
        p1_value = lumi.Variable(name="p1_value", type=lumi.VariableType.Scalar, sql="select 1 + 2")
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            variables=[p1_value],
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=p1_value,
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the parameter is added at the end of the options
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @@p1_value = select 1 + 2;
            @x = use Sys.Admin.SetupView with @@p1_value
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,@@p1_value,false,"p1 is nice"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_table_variable(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a table variable used as a
        # parameter
        p2_value = lumi.Variable(
            name="p2_value",
            type=lumi.VariableType.Table,
            sql="""select * from ( values(1, 2, 3), (4, 5, 6))""",
        )
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from @p2;",
            description="a view",
            variables=[p2_value],
            parameters=[
                lumi.Parameter(
                    name="p2",
                    type=lumi.ParameterType.Table,
                    value=p2_value,
                    is_mandatory=False,
                    tooltip="p2 is nice",
                )
            ],
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the parameter is added at the end of the options
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @p2_value = select * from ( values(1, 2, 3), (4, 5, 6));
            @x = use Sys.Admin.SetupView with @p2_value
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p2,Table,@p2_value,false,"p2 is nice"
            ----
            select * from @p2;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_two_variables(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a one table and one scalar
        # for a parameter
        p1_value = lumi.Variable(name="p1_value", type=lumi.VariableType.Scalar, sql="select 1 + 2")
        p2_value = lumi.Variable(
            name="p2_value",
            type=lumi.VariableType.Table,
            sql="""select * from ( values(1, 2, 3), (4, 5, 6))""",
        )
        # the table is used as a parameter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from @p2;",
            description="a view",
            variables=[p1_value, p2_value],
            parameters=[
                lumi.Parameter(
                    name="p2",
                    type=lumi.ParameterType.Table,
                    value=p2_value,
                    is_mandatory=False,
                    tooltip="p2 is nice",
                )
            ],
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the vars are initialised, the with statement includes them
        # and the table parameter is included
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @@p1_value = select 1 + 2;
            @p2_value = select * from ( values(1, 2, 3), (4, 5, 6));
            @x = use Sys.Admin.SetupView with @@p1_value, @p2_value
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p2,Table,@p2_value,false,"p2 is nice"
            ----
            select * from @p2;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_no_changes(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})])
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})]
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                )
            ]
        )

        # and a desired state that is the same;
        # new line at the end of script should not count as change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="""select * from foo
            """,
            description="a view",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and the read query was sent
        request = put.calls[0].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            select f.Content, r.Version from sys.file as f
            join sys.registration as r on r.Name = 'Views.Unit.Test'
            where path = 'databaseproviders/Views/Unit/Test.sql'
            order by r.Version asc
            limit 1
        """)
        assert request.content.decode() == expected
        # and we read the remote but didn't write anything
        assert len(respx_mock.calls) == 3

    def test_update_with_test_option_means_no_change(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        respx_mock.put("/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]).mock(
            side_effect=[httpx.Response(202, json={"executionId": exec_id})]
        )
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})]
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                )
            ]
        )

        # and a desired that is the same except we add test_options
        # limit, useDryRun, groupby, filter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            limit=42,
            use_dry_run=True,
            group_by="one",
            filter="something",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None because test_options are not saved
        assert state is None

    def test_update_with_extra_saved_option(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(
            # Read, update, registration check
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )

        # and a desired adds a saved option (documentationLink)
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="http://example.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we created a new view
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --documentationLink="http://example.com"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_does_registration_check(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(
            # Read, update, 2xregistration check
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[{"Version": 2}]),  # registration complete
            ]
        )

        # and a desired makes a change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="http://example.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and the registration check was performed twice
        # because the version before update was 1 and we wait until it's increased
        check2 = put.calls[3].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check2.content.decode("utf8") == expected

    def test_update_completes_when_registration_version_does_not_update(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(
            # Read, update, 3xregistration check
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 5
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 5
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                # register within the 3 tries
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
            ]
        )

        # and a desired makes a change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="http://example.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a state is returned even though the reg check didn't complete
        assert state == {"provider": "Views.Unit.Test"}
        # and the registration check was 3 times
        # because the version didn't update
        check = put.calls[4].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check.content.decode("utf8") == expected

    def test_update_completes_when_no_registration_rows_returned(self, respx_mock):
        # given an existing remote where the view is not going to
        # register within the 3 tries
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(
            # Read, update, 3xregistration check
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 5
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 5
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[]),  # no rows returned
            ]
        )

        # and a desired makes a change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="http://example.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a state is returned even though the reg check didn't complete
        assert state == {"provider": "Views.Unit.Test"}
        # and the registration check was 3 times
        # because the version didn't update
        check = put.calls[4].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check.content.decode("utf8") == expected

    def test_update_defaulted_saved_option(self, respx_mock):
        # given an existing View created without the variableShape option
        exec_id = "abcdefg"
        respx_mock.put("/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]).mock(
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [],
                                "IsWithinDirectProviderView": False,
                            },
                            version=1,
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )

        # and the desired leaves it empty
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo", description="a view"
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then no change is detected because False is a server default
        # and the user has not specified it
        assert state is None

    def test_update_with_modified_saved_option(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {"Description": "a view", "Parameters": [], "DocumentationLink": "bar.com"},
                            version=1,
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )

        # and a desired adds a saved option (documentationLink)
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="foo.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we created a new view
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --documentationLink="foo.com"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_modified_sql(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )
        # and a desired changes the sql
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from something_else",
            description="a view",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we created a new view
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            ----
            select * from something_else
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_extra_parameter(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )
        # and a desired that is the same but adds a parameter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=23,
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the new parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we update the view with the parameter
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,23,false,"p1 is nice"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_removed_parameter(self, respx_mock):
        # given an existing view with one parameter
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [{"Name": "existing", "Type": "Text"}],
                            },
                            version=1,
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )
        # and a desired that removes the parameter
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo", description="a view"
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we created a new view with none
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_modified_parameter(self, respx_mock):
        # given an existing view with integer parameter p1
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {"Description": "a view", "Parameters": [{"Name": "p1", "Type": "Int"}]},
                            version=1,
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 2}]),  # registration check
            ]
        )
        # and a desired that changes the parameter type to text
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Text,
                    value="some text",
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we modify the view with a text parameter
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Text,"some text",false,"p1 is nice"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_useasdefault_changes_value(self, respx_mock):
        # given an existing view with integer parameter p1
        # with a default of zero
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [
                                    {
                                        "Name": "p1",
                                        "Type": "Int",
                                        "DefaultValue": 0,
                                        "Description": "p1 is nice",
                                    }
                                ],
                            },
                            version=1,
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 2}]),  # registration check
            ]
        )
        # and a desired with setAsDefault=true which means to use the value as a default
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=20,
                    set_as_default_value=True,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we modify the view with the new value
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,20,true,"p1 is nice"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_useasdefault_same_value(self, respx_mock):
        # given an existing view with integer parameter p1
        # with a default of 20
        exec_id = "abcdefg"
        respx_mock.put("/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]).mock(
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [
                                    {
                                        "Name": "p1",
                                        "Type": "Int",
                                        "DefaultValue": 20,
                                        "Description": "p1 is nice",
                                    }
                                ],
                            },
                            version=1,
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 2}]),  # registration check
            ]
        )
        # and a desired with setAsDefault but a value also 20
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=20,
                    set_as_default_value=True,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then there is no change
        assert state is None

    def test_update_with_renamed_parameter(self, respx_mock):
        # given an existing view with integer parameter p1
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [{"Name": "p1", "Type": "Int", "ExtraField": "boo"}],
                            },
                            version=1,
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 2}]),  # registration check
            ]
        )
        # and a desired renames p1 to p2
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p2",
                    type=lumi.ParameterType.Int,
                    value=12,
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we modify the view to only have p2
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p2,Int,12,false,"p1 is nice"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_provider_name(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(200, json=[]),  # delete
                httpx.Response(200, json=[]),  # de-reg complete
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 1}]),  # reg complete
            ]
        )

        # given a view which has been created with a provider name
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we change the provider name
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.NotTheSame", sql="select * from foo", description="a view"
        )
        # and update
        sut.update(self.client, old_state)
        # then the existing provider gets deleted
        first = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider=Views.Unit.Test
            --deleteProvider
            ----
            select 1 as deleting
            enduse;
            select * from @x;
        """)
        assert first.content.decode() == expected
        # and a new one gets created
        request = put.calls[2].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.NotTheSame"
            --description="a view"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_no_changes_and_table_param(self, respx_mock):
        # given an existing remote

        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})])

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})]
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from #PARAMETERVALUE(p1);",
                            {
                                "Description": "a view",
                                "Parameters": [
                                    {
                                        "Columns": [
                                            {
                                                "IsPrimaryKey": "false",
                                                "IsMain": "false",
                                                "IsRequiredByProvider": "false",
                                                "MandatoryForActions": None,
                                                "ClientIds": None,
                                                "Name": "3 + 4",
                                                "Type": "BigInt",
                                                "Description": None,
                                                "DisplayName": "3 + 4",
                                                "ConditionUsage": 0,
                                                "SampleValues": None,
                                                "AllowedValues": None,
                                            }
                                        ],
                                        "Name": "p1",
                                        "Type": "Table",
                                        "Description": "I am a description"
                                        "\nAvailable columns:\n3 + 4 (BigInt)",
                                        "DisplayName": "p1",
                                        "ConditionUsage": 0,
                                    },
                                    {
                                        "Columns": [
                                            {
                                                "IsPrimaryKey": "false",
                                                "IsMain": "false",
                                                "IsRequiredByProvider": "false",
                                                "MandatoryForActions": None,
                                                "ClientIds": None,
                                                "Name": "5 + 6",
                                                "Type": "BigInt",
                                                "Description": None,
                                                "DisplayName": "5 + 6",
                                                "ConditionUsage": 0,
                                                "SampleValues": None,
                                                "AllowedValues": None,
                                            }
                                        ],
                                        "Name": "p2",
                                        "Type": "Table",
                                        "Description": "\nAvailable columns:\n5 + 6 (BigInt)",
                                        "DisplayName": "p2",
                                        "ConditionUsage": 0,
                                    },
                                ],
                            },
                            version=1,
                        )
                    ],
                )
            ]
        )

        p1_value = lumi.Variable(name="p1", type=lumi.VariableType.Table, sql="select * from 3 + 4")

        p2_value = lumi.Variable(name="p2", type=lumi.VariableType.Table, sql="select a,b,c from Test")
        # and a desired state that is the same;
        # new line at the end of script should not count as change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="""select * from #PARAMETERVALUE(p1);""",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    value=p1_value,
                    is_mandatory=False,
                    tooltip="I am a description",
                    type=lumi.ParameterType.Table,
                ),
                lumi.Parameter(
                    name="p2",
                    value=p2_value,
                    is_mandatory=False,
                    type=lumi.ParameterType.Table,
                    tooltip="",
                ),
            ],
        )

        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and the read query was sent
        request = put.calls[0].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
               select f.Content, r.Version from sys.file as f
               join sys.registration as r on r.Name = 'Views.Unit.Test'
               where path = 'databaseproviders/Views/Unit/Test.sql'
               order by r.Version asc
               limit 1
           """)
        assert request.content.decode() == expected
        # and we read the remote but didn't write anything
        assert len(respx_mock.calls) == 3

    @staticmethod
    def test_deps_none():
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo", description="a view"
        )
        assert sut.deps() == []

    @staticmethod
    def test_deps_extra():
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            dependencies=[1, 2, 3],
        )
        assert sut.deps() == [1, 2, 3]

    @staticmethod
    def test_deps_with_inline_resource():
        # given an inline resource
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        inline = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="NewProvider",  # Different provider
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        # when we make a view that depends on it (eg because it references some of the
        # columns)
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            dependencies=[inline],
        )
        # then the dependencies include the inline
        assert sut.deps() == [inline]

    def test_delete(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # delete
                httpx.Response(200, json=[]),  # delete
                httpx.Response(200, json=[]),  # de-registration check
            ]
        )
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="""select * from 2+2""", description="a view"
        )

        # given a view that exists in the remote
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we delete it
        sut.delete(self.client, old_state)
        # then sql to remove the view
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider=Views.Unit.Test
            --deleteProvider
            ----
            select 1 as deleting
            enduse;
            select * from @x;
        """)
        assert request.content.decode() == expected

    def test_delete_does_registration_check(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(200, json=[]),  # delete
                httpx.Response(200, json=[{"Version": 5}]),  # still registered
                httpx.Response(200, json=[]),  # deregistered
            ]
        )

        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="""select * from 2+2""", description="a view"
        )

        # given a view that exists in the remote
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we delete it
        sut.delete(self.client, old_state)
        # the registration check was performed twice because we waited for
        # the view to be removed
        check = put.calls[2].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check.content.decode("utf8") == expected

    def test_delete_completes_when_reg_incomplete(self, respx_mock, capsys):
        # given de-registration takes more than 3 tries
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(200, json=[]),  # delete
                httpx.Response(200, json=[{"Version": 5}]),  # still registered
                httpx.Response(200, json=[{"Version": 5}]),  # still registered
                httpx.Response(200, json=[{"Version": 5}]),  # still registered
            ]
        )

        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="""select * from 2+2""", description="a view"
        )
        # given a view that exists in the remnte
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we delete it
        sut.delete(self.client, old_state)
        # a delete was sent and the registration check was performed 3 times
        check = put.calls[3].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check.content.decode("utf8") == expected
        # check warning is returned
        captured = capsys.readouterr()
        assert captured.err == (
            f"warning: no view deregistration after "
            f"{lumi.ViewResource.Registration.tries}"
            f" tries for Views.Unit.Test"
        )

    def test_dump(self):
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            documentation_link="http://example.com/docs",
            variable_shape=True,
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=23,
                    set_as_default_value=True,
                    tooltip="Parameter 1"
                )
            ]
        )
        result = sut.model_dump(context={"style": "dump"})
        # id field is globally excluded, so it won't appear in dumps
        assert "id" not in result
        assert result["provider"] == "Views.Unit.Test"
        assert result["sql"] == "select * from foo;"
        assert result["description"] == "a view"
        assert result["documentation_link"] == "http://example.com/docs"
        assert result["variable_shape"] is True
        assert len(result["parameters"]) == 1
        assert result["parameters"][0]["name"] == "p1"
        assert result["parameters"][0]["type"] == "Int"
        assert result["parameters"][0]["value"] == 23

    def test_undump(self):
        data = {
            "provider": "Views.Unit.Test",
            "sql": "select * from foo;",
            "description": "a view for undump testing",
            "documentation_link": "http://example.com/docs",
            "variable_shape": True,
            "parameters": [
                {
                    "name": "p1",
                    "type": "Int",
                    "value": 23,
                    "set_as_default_value": True,
                    "tooltip": "Parameter 1"
                }
            ]
        }
        result = lumi.ViewResource.model_validate(data, context={"style": "dump", "id": "my-view"})
        assert result.id == "my-view"
        assert result.provider == "Views.Unit.Test"
        assert result.sql == "select * from foo;"
        assert result.description == "a view for undump testing"
        assert result.documentation_link == "http://example.com/docs"
        assert result.variable_shape is True
        assert len(result.parameters) == 1
        assert result.parameters[0].name == "p1"
        assert result.parameters[0].type == lumi.ParameterType.Int
        assert result.parameters[0].value == 23


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeBackoffHandler:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @staticmethod
    def test_pause_time_smaller_than_max_pause_time():
        backoff_handler_test = BackoffHandler(pause_time=0.5, max_pause_time=20, beta=1.0001)
        backoff_handler_test.sleep()
        pytest.approx(0.50005, backoff_handler_test.pause_time, 0.00001)

    @staticmethod
    def test_pause_time_equals_max_pause_time():
        backoff_handler_test = BackoffHandler(pause_time=2, max_pause_time=2, beta=1.001)
        backoff_handler_test.sleep()
        assert backoff_handler_test.pause_time == 2

    @staticmethod
    def test_pause_time_bigger_than_max_pause_time():
        with pytest.raises(ValueError) as ex:
            BackoffHandler(pause_time=100, max_pause_time=20, beta=1.1)
            assert "Pause time must be between 0.1 and 20, both inclusive." in str(ex.value)

    @staticmethod
    def test_default_optional_values():
        backoff_handler_test = BackoffHandler()
        backoff_handler_test.sleep()
        pytest.approx(0.1005012520859401, backoff_handler_test.pause_time, 0.00001)

    @staticmethod
    def test_call_1000_times_pause_time_equals_max_pause_time():
        backoff_handler_test = BackoffHandler()
        # I have checked that this is constant after 922 requests
        for i in range(1000):
            backoff_handler_test._update_pause_time()
        assert backoff_handler_test.pause_time == 10


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeQuery:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_wait_for_background_switch_case_correct(self, respx_mock):
        exec_id = "abcdefg"
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                httpx.Response(200, json={"status": TaskStatus.WAITING_FOR_ACTIVATION}),
                httpx.Response(200, json={"status": TaskStatus.WAITING_TO_RUN}),
                httpx.Response(200, json={"status": TaskStatus.CREATED}),
                httpx.Response(200, json={"status": TaskStatus.WAITING_FOR_CHILDREN_TO_COMPLETE}),
                httpx.Response(200, json={"status": TaskStatus.RUNNING}),
                httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION}),
            ]
        )
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is True

    @pytest.mark.parametrize("status", [TaskStatus.FAULTED, TaskStatus.CANCELED])
    def test_wait_for_background_status_faulted_throws(self, respx_mock, status):
        exec_id = "abcdefg"
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                httpx.Response(200, json={"status": TaskStatus.WAITING_FOR_ACTIVATION}),
                httpx.Response(200, json={"status": status}),
            ]
        )
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        with pytest.raises(RuntimeError) as ex:
            lumi.wait_for_background(client=self.client, execution_id=exec_id)
        assert str(ex.value) == f"Query was {status}. Execution id: abcdefg"

    def test_wait_for_background_status_unknown_throws(self, respx_mock):
        exec_id = "abcdefg"
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                httpx.Response(200, json={"status": TaskStatus.WAITING_FOR_ACTIVATION}),
                httpx.Response(200, json={"status": "someunknownstatus"}),
            ]
        )
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        with pytest.raises(RuntimeError) as ex:
            lumi.wait_for_background(client=self.client, execution_id=exec_id)
        assert str(ex.value) == "Unknown status: someunknownstatus. Execution id: abcdefg"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeInlinePropertiesResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture()
    def default_content(self):
        return [{
            "Content": """Key,Name,DataType,Description,IsMain,AsAt\n"""
                       """Domain/Scope/Code,PropName,Boolean,A description,false,""",
            "Error": "",
        }]

    def test_create_no_properties(self, respx_mock: Router, default_content):
        setup_sql_background_routes(respx_mock, result_data=default_content)
        # given an inline properties resource with zero properties
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
            ],
        )
        # when we create it
        sut.create(self.client)
        # then the request sql creates the provider extension but with no properties
        expected = textwrap.dedent("""\
            select * from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'TestProvider'
                and WriteAction = 'Set'\n
            ;
        """)
        lumi_query = respx_mock.calls[0].request.content
        assert lumi_query.decode() == expected

    def test_create_basic_properties(self, respx_mock: Router, default_content):
        setup_sql_background_routes(respx_mock, result_data=default_content)
        # given an inline properties resource with basic properties
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        # then the request sql matches
        expected = textwrap.dedent("""\
            @keysToCatalog = values
                ('Instrument/TestScope/TestCode', 'Test Property', 'Text', 'A test property', null, null)
            ;
            @config = select
                column1 as [Key], column2 as Name, column3 as DataType, column4 as Description, column5 as IsMain, column6 as AsAt
                from @keysToCatalog
            ;
            select * from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'TestProvider'
                and Configuration = @config
                and WriteAction = 'Set'\n
            ;
        """)  # noqa: E501
        lumi_query = respx_mock.calls[0].request.content
        assert lumi_query.decode() == expected
        # and the state is returned
        assert state == {
            "provider": "TestProvider",
            "source_version": "534a9af588cf7b5315bd3d4e88d3c82e5c2cf8d271a86490c13496eb7745a417",
            "remote_version": "8b16bcb769741f61d25e92f4966ea3694fd643bd7fdb7d11c72ad7ade7ca05dd",
            "extension": None
        }

    def test_create_with_asat(self, respx_mock: Router, default_content):
        setup_sql_background_routes(respx_mock, result_data=default_content)
        # given an inline properties resource with basic properties
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                    as_at=datetime(2025, 2, 3, 15, 41),
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        # then the request sql matches
        expected = textwrap.dedent("""\
            @keysToCatalog = values
                ('Instrument/TestScope/TestCode', 'Test Property', 'Text', 'A test property', null, '2025-02-03 15:41:00')
            ;
            @config = select
                column1 as [Key], column2 as Name, column3 as DataType, column4 as Description, column5 as IsMain, column6 as AsAt
                from @keysToCatalog
            ;
            select * from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'TestProvider'
                and Configuration = @config
                and WriteAction = 'Set'\n
            ;
        """)  # noqa
        lumi_query = respx_mock.calls[0].request.content
        assert lumi_query.decode() == expected
        # and the state is returned
        assert state == {
            "provider": "TestProvider",
            "source_version": "bc34d5f5a40b4cbb20dd2035cbe849f7c75f81210657035a9bb108290e3512ad",
            "remote_version": "8b16bcb769741f61d25e92f4966ea3694fd643bd7fdb7d11c72ad7ade7ca05dd",
            "extension": None
        }

    def test_create_with_provider_extension(self, respx_mock, default_content):
        setup_sql_background_routes(respx_mock, result_data=default_content)
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        # given an inline properties resource with provider extension
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            provider_name_extension="Extension",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        # then the state includes the extension
        assert state is not None
        assert state["extension"] == "Extension"
        # and the request sql matches includes extension
        expected = textwrap.dedent("""\
            @keysToCatalog = values
                ('Instrument/TestScope/TestCode', 'Test Property', 'Text', 'A test property', null, null)
            ;
            @config = select
                column1 as [Key], column2 as Name, column3 as DataType, column4 as Description, column5 as IsMain, column6 as AsAt
                from @keysToCatalog
            ;
            select * from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'TestProvider'
                and Configuration = @config
                and WriteAction = 'Set'
                and ProviderNameExtension = 'Extension'
            ;
        """)  # noqa: E501
        lumi_query = respx_mock.calls[0].request.content
        assert lumi_query.decode() == expected

    def test_create_multiple_properties_with_optional_fields(self, respx_mock, default_content):
        setup_sql_background_routes(respx_mock, result_data=default_content)
        # given two properties to inline where one isMain and AsAt
        prop_key1 = DefinitionRef(id="key1", domain=Domain.Instrument, scope="Scope1", code="Code1")
        prop_key2 = DefinitionRef(id="key2", domain=Domain.Portfolio, scope="Scope2", code="Code2")
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key1,
                    name="Property 1",
                    data_type=lumi.InlineDataType.Text,
                    description="First property",
                    is_main=True,
                    as_at=datetime(2023, 1, 1, 12, 0, 0),
                ),
                lumi.InlineProperty(
                    key=prop_key2,
                    name="Property 2",
                    data_type=lumi.InlineDataType.Decimal,
                    description="Second property",
                    is_main=False,
                ),
            ],
        )
        # when we create it
        sut.create(self.client)
        # then the SQL includes all columns and values
        expected = textwrap.dedent("""\
            @keysToCatalog = values
                ('Instrument/Scope1/Code1', 'Property 1', 'Text', 'First property', true, '2023-01-01 12:00:00'),('Portfolio/Scope2/Code2', 'Property 2', 'Decimal', 'Second property', false, null)
            ;
            @config = select
                column1 as [Key], column2 as Name, column3 as DataType, column4 as Description, column5 as IsMain, column6 as AsAt
                from @keysToCatalog
            ;
            select * from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'TestProvider'
                and Configuration = @config
                and WriteAction = 'Set'\n
            ;
        """)  # noqa
        lumi_query = respx_mock.calls[0].request.content
        assert lumi_query.decode() == expected

    def test_create_custom_entity(self, respx_mock, default_content):
        setup_sql_background_routes(respx_mock, result_data=default_content)
        # given an inline properties resource for a custom entity
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    name="Test Property",
                    field="cefield",
                    data_type=lumi.InlineDataType.Text,
                    description="A test field",
                    as_at=datetime(2025, 2, 3, 15, 41),
                )
            ],
        )
        # when the sql is generated
        sut.create(self.client)
        # then the property key in the sql is just the field
        expected = textwrap.dedent("""\
            @keysToCatalog = values
                ('cefield', 'Test Property', 'Text', 'A test field', null, '2025-02-03 15:41:00')
            ;
            @config = select
                column1 as [Key], column2 as Name, column3 as DataType, column4 as Description, column5 as IsMain, column6 as AsAt
                from @keysToCatalog
            ;
            select * from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'TestProvider'
                and Configuration = @config
                and WriteAction = 'Set'\n
            ;
        """)  # noqa
        lumi_query = respx_mock.calls[0].request.content
        assert lumi_query.decode() == expected

    def test_create_sql_structured_property(self, respx_mock, default_content):
        setup_sql_background_routes(respx_mock, result_data=default_content)
        # given an inline properties resource for a property which is structured (non scalar)
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    field="subfield",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                    as_at=datetime(2025, 2, 3, 15, 41),
                )
            ],
        )
        # when the sql is generated
        sut.create(self.client)
        # then the property key is the domain/scope/code.field
        expected = textwrap.dedent("""\
            @keysToCatalog = values
                ('Instrument/TestScope/TestCode.subfield', 'Test Property', 'Text', 'A test property', null, '2025-02-03 15:41:00')
            ;
            @config = select
                column1 as [Key], column2 as Name, column3 as DataType, column4 as Description, column5 as IsMain, column6 as AsAt
                from @keysToCatalog
            ;
            select * from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'TestProvider'
                and Configuration = @config
                and WriteAction = 'Set'\n
            ;
        """)  # noqa
        lumi_query = respx_mock.calls[0].request.content
        assert lumi_query.decode() == expected

    def make_csv(self, dict_data):
        output = io.StringIO()
        writer = csv.DictWriter(output, fieldnames=dict_data.keys())
        writer.writeheader()
        writer.writerow(dict_data)
        return output.getvalue()

    def test_read_existing_properties(self, respx_mock):  # noqa: E501
        # given an instrument property is already inlined
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        # mock the csv response from lumi
        existing = {
            "Key": "Instrument/TestScope/TestCode",
            "Name": "Test Property",
            "DataType": "Text",
            "Description": "A test property"
        }
        csv_content = self.make_csv(existing)
        setup_sql_background_routes(respx_mock, result_data=[{"Content": csv_content, "Error": ""}])
        # when we read it
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        old_state = SimpleNamespace(provider="TestProvider", extension=None)
        result = sut.read(self.client, old_state)
        # then we get the parsed properties keyed by the property
        assert result == {
            "Instrument/TestScope/TestCode": {
                "AsAt": None,
                "Name": "Test Property",
                "DataType": "Text",
                "Description": "A test property",
                "IsMain": None
            }
        }
        # and the SQL request to read is right
        request = respx_mock.calls[0].request
        actual = request.content.decode("utf8")
        expected = textwrap.dedent("""\
            select Content from Sys.File
                where Path = 'config/lusid/factories/testproviderproviderfactory.csv'
            ;
        """)  # noqa
        assert actual == expected

    def test_read_existing_with_extension(self, respx_mock):  # noqa: E501
        # given an instrument property is already inlined
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        # mock the csv response from lumi
        existing = {
            "Key": "Instrument/TestScope/TestCode",
            "Name": "Test Property",
            "DataType": "Text",
            "Description": "A test property"
        }
        csv_content = self.make_csv(existing)
        setup_sql_background_routes(respx_mock, result_data=[{"Content": csv_content}])
        # when we read it
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        old_state = SimpleNamespace(provider="TestProvider", extension="Ext")
        result = sut.read(self.client, old_state)
        # then we get the parsed properties keyed by the property
        assert result == {
            "Instrument/TestScope/TestCode": {
                "AsAt": None,
                "Name": "Test Property",
                "DataType": "Text",
                "Description": "A test property",
                "IsMain": None
            }
        }
        # and path in the query, inserts the extension into the factory name
        request = respx_mock.calls[0].request
        actual = request.content.decode("utf8")
        expected = textwrap.dedent("""\
            select Content from Sys.File
                where Path = 'config/lusid/factories/ext_testproviderproviderfactory.csv'
            ;
        """)  # noqa
        assert actual == expected

    def test_update_no_changes(self, respx_mock):
        # given an instrument property is already inlined
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        # mock the csv response from lumi
        existing = {
            "Key": "Instrument/TestScope/TestCode",
            "Name": "Test Property",
            "DataType": "Text",
            "Description": "A test property"
        }
        csv_content = self.make_csv(existing)
        setup_sql_background_routes(respx_mock, result_data=[{"Content": csv_content}])
        # and a desired state with the same content hash
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        # create matching state hashes
        desired = sut.model_dump(mode="json", by_alias=True, exclude_none=True)
        source_version = hashlib.sha256(json.dumps(desired, sort_keys=True).encode()).hexdigest()
        remote_version = hashlib.sha256(
            json.dumps({
                "Instrument/TestScope/TestCode": {
                    "Name": "Test Property",
                    "DataType": "Text",
                    "Description": "A test property",
                    "IsMain": None,
                    "AsAt": None,
                }
            }, sort_keys=True,
            ).encode()
        ).hexdigest()
        old_state = SimpleNamespace(
            source_version=source_version,
            remote_version=remote_version,
            provider="TestProvider",
            extension=None,
        )
        # when we update
        result = sut.update(self.client, old_state)
        # then no update is performed
        assert result is None

    def test_update_with_changes(self, respx_mock):
        # given an instrument property is already inlined
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        #     mock the csv response from lumi
        existing = {
            "Key": "Instrument/TestScope/TestCode",
            "Name": "Test Property",
            "DataType": "Text",
            "Description": "Old Description"
        }
        csv_content = self.make_csv(existing)
        setup_sql_background_routes(respx_mock, result_data=[{"Content": csv_content, "Error": ""}])
        #     mock the read
        routes = setup_sql_background_routes(respx_mock, result_data=[
            {"Content": csv_content, "Error": ""}
        ])
        #     mock the create
        exec_id = "abcdefg"
        routes["put"].mock(
            side_effect=[
                httpx.Response(202, json={"executionId": exec_id}),  # read
                httpx.Response(202, json={"executionId": exec_id}),  # create
            ]
        )
        routes["result"].mock(
            side_effect=[
                httpx.Response(200, json=[{"Content": csv_content}]),  # read result
                httpx.Response(200, json=[{"Content": csv_content, "Error": ""}]),  # create result
            ]
        )
        routes["status"].mock(
            side_effect=[
                httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION}),  # read
                httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION}),  # create
            ]
        )
        # when we rename the column and change the description
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Updated Property",  # Changed name
                    data_type=lumi.InlineDataType.Text,
                    description="Updated description",  # Changed description
                )
            ],
        )
        old_state = SimpleNamespace(
            source_version="old_version",
            remote_version="old_remote_version",
            provider="TestProvider",
            extension=None,
        )
        result = sut.update(self.client, old_state)
        # then create is called and state is returned
        assert result is not None
        assert result["provider"] == "TestProvider"

    def test_update_provider_change_triggers_delete(self, respx_mock):
        exec_id = "abcdefg"
        existing = {
            "Key": "Instrument/TestScope/TestCode",
            "Name": "Test Property",
            "DataType": "Text",
            "Description": "Old Description"
        }
        csv_content = self.make_csv(existing)
        # Setup for read + delete + create operations (3 total)
        respx_mock.put("/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]).mock(
            return_value=httpx.Response(202, json={"executionId": exec_id})
        )
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            return_value=httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION}),  # read
        )
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
                return_value=httpx.Response(200, json=[{"Content": csv_content, "Error": ""}]),
        )
        # given an existing inline
        old_state = SimpleNamespace(
            source_version="old_version",
            remote_version="old_remote_version",
            provider="OldProvider",  # Different from current
            extension=None,
        )
        # when we update the desired with a new provider name
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="NewProvider",  # Different provider
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        new_state = sut.update(self.client, old_state)
        # then delete is called first, then create
        del_request = respx_mock.calls[0].request.content.decode()
        assert del_request == textwrap.dedent("""\
            select *
                from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'OldProvider'
                and WriteAction = 'Reset'\n
            ;
        """)
        assert new_state == {
            "provider": "NewProvider",
            "extension": None,
            "remote_version": "323b73ace7abb1e03a9b50166d725ede4945ab215736970e5afba7165f624d20",
            "source_version": "6a92efc7df690b538e5d530e2b4808c842e85fbe84a5b5ee88cf8a4a4bf7f0ae"
        }

    def test_update_extension_calls_delete(self, respx_mock):
        exec_id = "abcdefg"
        #     mock the csv response from lumi
        existing = {
            "Key": "Instrument/TestScope/TestCode",
            "Name": "Test Property",
            "DataType": "Text",
            "Description": "Old Description"
        }
        csv_content = self.make_csv(existing)
        # Setup for read + delete + create operations (3 total)
        respx_mock.put("/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]).mock(
            return_value=httpx.Response(202, json={"executionId": exec_id})
        )
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            return_value=httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION}),  # read
        )
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
                return_value=httpx.Response(200, json=[{"Content": csv_content, "Error": ""}]),
        )
        # given an existing inline
        old_state = SimpleNamespace(
            source_version="old_version",
            remote_version="old_remote_version",
            provider="OldProvider",  # Different from current
            extension=None,
        )
        # when we update the extension
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="OldProvider",
            provider_name_extension="NewExt",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        new_state = sut.update(self.client, old_state)
        # then delete is called first, then create
        del_request = respx_mock.calls[0].request.content.decode()
        assert del_request == textwrap.dedent("""\
            select *
                from Sys.Admin.Lusid.Provider.Configure
                where Provider = 'OldProvider'
                and WriteAction = 'Reset'\n
            ;
        """)
        assert new_state == {
            "provider": "OldProvider",
            "extension": "NewExt",
            "remote_version": "323b73ace7abb1e03a9b50166d725ede4945ab215736970e5afba7165f624d20",
            "source_version": "e57c2d65ac17bda7b0f93f1eaadeecb204595530fa665b49f920720001d6a52a"
        }

    def test_delete(self, respx_mock):
        setup_sql_background_routes(respx_mock, result_data=[])
        # given an existing inline
        old_state = SimpleNamespace(
            source_version="old_version",
            remote_version="old_remote_version",
            provider="OldProvider",  # Different from current
            extension="extension",
        )
        # when we delete it
        lumi.InlinePropertiesResource.delete(self.client, old_state)
        # then the writeaction is reset which removes the inline
        del_request = respx_mock.calls[0].request.content.decode()
        expected = textwrap.dedent("""\
           select *
               from Sys.Admin.Lusid.Provider.Configure
               where Provider = 'OldProvider'
               and WriteAction = 'Reset'
               and ProviderNameExtension = 'extension'
           ;
        """)
        assert expected == del_request

    def test_deps_lusid_provider(self):
        # given an inline properties resource with basic properties
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        # when we get the dependencies the property definition is returned
        assert sut.deps() == [prop_key]

    def test_deps_custom_entity(self):
        # given an inline properties resource for a custom entity provider
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    field="action",
                    name="Test Property",
                    data_type=lumi.InlineDataType.Text,
                    description="A test property",
                )
            ],
        )
        # when we get the dependencies there are none
        assert sut.deps() == []

    def test_dump(self):
        # given a inline prop
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        sut = lumi.InlinePropertiesResource(
            id="test-inline-props",
            provider="TestProvider",
            properties=[
                lumi.InlineProperty(
                    key=prop_key,
                    name="Test Property",
                    data_type=lumi.InlineDataType.Boolean,
                    description="A test property",
                )
            ],
        )
        # when we dump it
        dump = sut.model_dump(by_alias=True, round_trip=True, exclude_none=True, context={
            "style": "dump"
        })
        # then everything gets dumped out and the property is ref'd
        assert dump == {
           "provider": "TestProvider",
           "properties": [
               {
                   "dataType": lumi.InlineDataType.Boolean,
                   "description": "A test property",
                   "key": {"$ref": "test-key"},
                   "name": "Test Property",
               },
           ],
       }

    def test_undump(self):
        # given a dump
        prop_key = DefinitionRef(
            id="test-key", domain=Domain.Instrument, scope="TestScope", code="TestCode"
        )
        dump = {
           "provider": "TestProvider",
           "properties": [
               {
                   "dataType": "Text",
                   "description": "A test property",
                   "key": {"$ref": "test-key"},
                   "name": "Test Property",
               },
           ],
        }
        # when we undump
        sut = lumi.InlinePropertiesResource.model_validate(dump, context={
            "id": "something-unique",
            "$refs": {
                "test-key": prop_key
            }
        })
        # then the id is extracted from the context
        assert sut.id == "something-unique"
        # and the property reference has been wired in
        assert sut.properties[0].key == prop_key



import pathlib

from fbnconfig import Deployment, drive


def configure(env):
    deployment_name = getattr(env, "name", "folders")
    base_folder = getattr(env, "base_dir", "folder")

    f1 = drive.FolderResource(id="base_folder", name=base_folder, parent=drive.root)
    f2 = drive.FolderResource(id="sub_folder", name="subfolder", parent=f1)
    f3 = drive.FolderResource(id="sub_sub_folder", name="subfolder2", parent=f2)
    content_path = pathlib.Path(__file__).parent.resolve() / pathlib.Path("poem.txt")
    ff = drive.FileResource(id="file1", folder=f3, name="myfile.txt", content_path=content_path)
    return Deployment(deployment_name, [f1, f2, f3, ff])


if __name__ == "__main__":
    import os

    import click

    import fbnconfig

    @click.command()
    @click.argument("lusid_url", envvar="LUSID_ENV", type=str)
    @click.option("-v", "--vars_file", type=click.File("r"))
    def cli(lusid_url, vars_file):
        host_vars = fbnconfig.load_vars(vars_file)
        d = configure(host_vars)
        fbnconfig.deploy(d, lusid_url, os.environ["FBN_ACCESS_TOKEN"])

    cli()



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import workspace

TEST_BASE = "https://foo.lusid.com"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWorkspaceResource:
    client = httpx.Client(
        base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]}
    )

    def test_create(self, respx_mock):
        respx_mock.post("/api/api/workspaces/personal").mock(return_value=httpx.Response(200, json={}))
        # given no workspace exists
        sut = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.PERSONAL,
            description="a test workspace",
            name="workspace1",
        )
        # when we call create
        new_state = sut.create(self.client)
        # then a post request is made
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/workspaces/personal"
        assert json.loads(request.content) == {
            "name": "workspace1",
            "description": "a test workspace",
        }
        # and the new state is returned
        assert new_state == {
            "visibility": "personal",
            "name": "workspace1",
            "content_hash": "55b0645f9146503f4bf66d88b85603a2ea8081b1272f2a1c1fb7bae84a0c209d",
        }

    def test_create_conflict(self, respx_mock):
        respx_mock.post("/api/api/workspaces/personal").mock(return_value=httpx.Response(429, json={}))
        # given a workspace with this name already exists
        sut = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.PERSONAL,
            description="a test workspace",
            name="workspace1",
        )
        # when we call create we get an exception
        with pytest.raises(httpx.HTTPError):
            sut.create(self.client)

    def test_delete(self, respx_mock):
        respx_mock.delete("/api/api/workspaces/personal/workspace1").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        # given an existing workspace
        # when we delete it
        old_state = SimpleNamespace(visibility="personal", name="workspace1", content_hash="xxxx")
        workspace.WorkspaceResource.delete(self.client, old_state)
        # then a delete request is made
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/api/api/workspaces/personal/workspace1"

    def test_deps(self):
        # given a workspace
        sut = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.PERSONAL,
            description="a test workspace",
            name="workspace1",
        )
        # it has no deps
        assert sut.deps() == []

    def test_read_exists(self, respx_mock):
        # given the workspace exists
        respx_mock.get("/api/api/workspaces/shared/workspace1").mock(
            return_value=httpx.Response(200, json={
                "name": "workspace1",
                "description": "a test workspace",
                "version": {"stuff": 1},
                "links": {"stuff": 2},
            })
        )
        sut = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.PERSONAL,
            description="not the same",
            name="workspace1",
        )
        # when we read it
        old_state = SimpleNamespace(visibility="shared", name="workspace1")
        res = sut.read(self.client, old_state)
        # then the links and version fields are removed
        assert res == {
            "name": "workspace1",
            "description": "a test workspace",
        }

    def test_update_same_hash(self):
        # given the desired resource has not changed (same hash as last time)
        sut = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.PERSONAL,
            description="a test workspace",
            name="workspace1",
        )
        old_state = SimpleNamespace(
            visibility="personal",
            name="workspace1",
            content_hash="55b0645f9146503f4bf66d88b85603a2ea8081b1272f2a1c1fb7bae84a0c209d",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then the hash check matches, no request is made and None is returned
        assert new_state is None

    def test_update_description(self, respx_mock):
        respx_mock.put("/api/api/workspaces/personal/workspace1").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the desired resource has modified only the description
        sut = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.PERSONAL,
            description="a new description",
            name="workspace1",
        )
        old_state = SimpleNamespace(
            visibility="personal",
            name="workspace1",
            content_hash="55b0645f9146503f4bf66d88b85603a2ea8081b1272f2a1c1fb7bae84a0c209d",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then a put request is made to update the description
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/api/api/workspaces/personal/workspace1"
        assert json.loads(request.content) == {
            "name": "workspace1",
            "description": "a new description",
        }
        # and the new state is returned with a new hash
        assert new_state == {
            "visibility": "personal",
            "name": "workspace1",
            "content_hash": "c69f472de0c5e825760d4b216e321454028433fe9ee41d584f171b5a7ea3b1c2",
        }

    def test_update_name(self, respx_mock):
        respx_mock.delete("/api/api/workspaces/personal/workspace1").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        respx_mock.post("/api/api/workspaces/personal").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the desired name has changed from workspace1 to workspace2
        sut = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.PERSONAL,
            description="a test workspace",
            name="workspace2",
        )
        old_state = SimpleNamespace(
            visibility="personal",
            name="workspace1",
            content_hash="55b0645f9146503f4bf66d88b85603a2ea8081b1272f2a1c1fb7bae84a0c209d",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then a post request creates the workspace with a new name
        request1 = respx_mock.calls[0].request
        assert request1.method == "POST"
        assert request1.url.path == "/api/api/workspaces/personal"
        assert json.loads(request1.content) == {
            "name": "workspace2",
            "description": "a test workspace",
        }
        # and a delete request is made to remove the existing name
        request2 = respx_mock.calls[1].request
        assert request2.method == "DELETE"
        assert request2.url.path == "/api/api/workspaces/personal/workspace1"
        # and the new state is returned with a new hash
        assert new_state == {
            "visibility": "personal",
            "name": "workspace2",
            "content_hash": "8a6fee24a682ea3d6995a93fdc1f2caabcc0b2e0bb9bcc90195aef036ff8406e",
        }

    def test_update_visibility(self, respx_mock):
        respx_mock.delete("/api/api/workspaces/personal/workspace1").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        respx_mock.post("/api/api/workspaces/shared").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the desired visibility has changed from personal to shared
        sut = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            description="a test workspace",
            name="workspace1",
        )
        old_state = SimpleNamespace(
            visibility="personal",
            name="workspace1",
            content_hash="55b0645f9146503f4bf66d88b85603a2ea8081b1272f2a1c1fb7bae84a0c209d",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then a post request creates the workspace in shared
        request1 = respx_mock.calls[0].request
        assert request1.method == "POST"
        assert request1.url.path == "/api/api/workspaces/shared"
        assert json.loads(request1.content) == {
            "name": "workspace1",
            "description": "a test workspace",
        }
        # then a delete request is made to remove the existing name
        request2 = respx_mock.calls[1].request
        assert request2.method == "DELETE"
        assert request2.url.path == "/api/api/workspaces/personal/workspace1"
        # and the new state is returned with the same hash because content hasn't changed
        assert new_state == {
            "visibility": "shared",
            "name": "workspace1",
            "content_hash": "55b0645f9146503f4bf66d88b85603a2ea8081b1272f2a1c1fb7bae84a0c209d",
        }

    def test_dump_simple_workspace(self):
        # given a simple workspace resource
        sut = workspace.WorkspaceResource(
            id="dump-workspace",
            visibility=workspace.Visibility.PERSONAL,
            name="DumpWorkspace",
            description="A test workspace for dumping"
        )
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then all fields are included (no excludes)
        expected = {
            "visibility": "personal",
            "name": "DumpWorkspace",
            "description": "A test workspace for dumping"
        }
        assert result == expected

    def test_undump_simple_workspace(self):
        # given dump data
        data = {
            "visibility": "shared",
            "name": "UndumpWorkspace",
            "description": "A test workspace for undumping"
        }
        # when we undump it with id from context
        result = workspace.WorkspaceResource.model_validate(
            data, context={"style": "dump", "id": "workspace_id"}
        )
        # then it's correctly populated including id from context
        assert result.id == "workspace_id"
        assert result.visibility == workspace.Visibility.SHARED
        assert result.name == "UndumpWorkspace"
        assert result.description == "A test workspace for undumping"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWorkspaceRef:
    client = httpx.Client(
        base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]}
    )

    def test_attach_when_present(self, respx_mock):
        # given that the remote exists
        respx_mock.get("/api/api/workspaces/shared/workspace1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = workspace.WorkspaceRef(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            name="workspace1",
        )
        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/api/api/workspaces/shared/workspace1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        sut = workspace.WorkspaceRef(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            name="workspace1",
        )
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Workspace shared/workspace1 not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/api/api/workspaces/shared/workspace1").mock(
            return_value=httpx.Response(400, json={})
        )
        client = self.client
        sut = workspace.WorkspaceRef(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            name="workspace1",
        )
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWorkspaceItemRef:
    client = httpx.Client(
        base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]}
    )

    def test_attach_with_workspace_resource(self, respx_mock):
        # given that the remote exists
        respx_mock.get("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        wksp = workspace.WorkspaceResource(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            name="workspace1",
            description="a test workspace",
        )
        sut = workspace.WorkspaceItemRef(
            id="wk1",
            name="item1",
            group="group1",
            workspace=wksp,
        )
        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_with_workspace_ref(self, respx_mock):
        # given that the remote exists
        respx_mock.get("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        wksp = workspace.WorkspaceRef(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            name="workspace1",
        )
        sut = workspace.WorkspaceItemRef(
            id="wk1",
            name="item1",
            group="group1",
            workspace=wksp,
        )
        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        wksp = workspace.WorkspaceRef(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            name="workspace1",
        )
        sut = workspace.WorkspaceItemRef(
            id="wk1",
            name="item1",
            group="group1",
            workspace=wksp,
        )
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Workspace item shared/workspace1/items/group1/item1 not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(400, json={})
        )
        client = self.client
        wksp = workspace.WorkspaceRef(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            name="workspace1",
        )
        sut = workspace.WorkspaceItemRef(
            id="wk1",
            name="item1",
            group="group1",
            workspace=wksp,
        )
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWorkspaceItemResource:
    client = httpx.Client(
        base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]}
    )

    @pytest.fixture
    def workspace1(self):
        return workspace.WorkspaceRef(
            id="wk1",
            visibility=workspace.Visibility.SHARED,
            name="workspace1",
        )

    @pytest.fixture
    def workspace2(self):
        return workspace.WorkspaceRef(
            id="wk1",
            visibility=workspace.Visibility.PERSONAL,
            name="workspace2",
        )

    def test_create(self, respx_mock, workspace1):
        respx_mock.post("/api/api/workspaces/shared/workspace1/items").mock(
            return_value=httpx.Response(200, json={})
        )
        # given a workspace but no item1
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            group="group1",
            format=2,
            description="a test item",
            name="item1",
            content={"key1": "value1", "key2": True, "key3": 42}
        )
        # when we call create
        new_state = sut.create(self.client)
        # then a post request is made
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/workspaces/shared/workspace1/items"
        assert json.loads(request.content) == {
            "name": "item1",
            "description": "a test item",
            "format": 2,
            "group": "group1",
            "content": {"key1": "value1", "key2": True, "key3": 42},
            "type": "type1",
        }
        # and the new state is returned
        assert new_state == {
            "visibility": "shared",
            "name": "item1",
            "group": "group1",
            "workspace_name": "workspace1",
            "content_hash": "d7875024b7d17597c4779da2097b221b73fd30d86891376f3762b7433ecbedfa",
        }

    def test_create_conflict(self, respx_mock, workspace1):
        # given an item already exists
        respx_mock.post("/api/api/workspaces/shared/workspace1/items").mock(
            return_value=httpx.Response(429, json={})
        )
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            group="group1",
            format=2,
            description="a test item",
            name="item1",
            content={"key1": "value1", "key2": True, "key3": 42}
        )
        # when we call create we get an exception
        with pytest.raises(httpx.HTTPError):
            sut.create(self.client)

    def test_delete(self, respx_mock):
        respx_mock.delete("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        # given an existing item
        old_state = SimpleNamespace(
            name="item1",
            group="group1",
            visibility="shared",
            workspace_name="workspace1",
            content_hash="xxxx"
        )
        # when we delete it
        workspace.WorkspaceItemResource.delete(self.client, old_state)
        # then a delete request is made
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/api/api/workspaces/shared/workspace1/items/group1/item1"

    def test_deps_ref(self):
        # given an item that depends on a workspace ref
        ref = workspace.WorkspaceRef(
            id="ref1",
            visibility=workspace.Visibility.PERSONAL,
            name="ref1",
        )
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=ref,
            type="type1",
            group="group1",
            format=2,
            description="a test item",
            name="item1",
            content={"key1": "value1", "key2": True, "key3": 42}
        )
        # it's deps include the workspace
        assert sut.deps() == [ref]

    def test_deps_workspace(self, workspace1):
        # given an item that depends on a workspace
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            group="group1",
            format=2,
            description="a test item",
            name="item1",
            content={"key1": "value1", "key2": True, "key3": 42}
        )
        # it's deps include the workspace
        assert sut.deps() == [workspace1]

    def test_read_exists(self, respx_mock, workspace2):
        # given the workspace tem exists
        respx_mock.get("/api/api/workspaces/personal/workspace2/items/group1/item1").mock(
            return_value=httpx.Response(200, json={
                "name": "item1",
                "group": "group1",
                "format": 2,
                "content": {},
                "description": "a test item",
                "version": {"stuff": 1},
                "links": {"stuff": 2},
            })
        )
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace2,
            type="type1",
            group="group1",
            format=2,
            description="a test item",
            name="item1",
            content={}
        )
        # when we read it
        old_state = SimpleNamespace(
            visibility="personal",
            name="item1",
            workspace_name="workspace2",
            group="group1",
        )
        res = sut.read(self.client, old_state)
        # then the links and version fields are removed
        assert res == {
            "name": "item1",
            "group": "group1",
            "description": "a test item",
            "format": 2,
            "content": {},
        }

    def test_update_same_hash(self, workspace1):
        # given the desired resource has not changed (same hash as last time)
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            description="a test item",
            name="item1",
            group="group1",
            format=3,
            content={"key1": "value1", "key2": True, "key3": 42},
        )
        old_state = SimpleNamespace(
            visibility=workspace1.visibility,
            name="item1",
            group="group1",
            workspace_name=workspace1.name,
            content_hash="21368e5f74f8fa8651d55f88d48ba651e3e8f0c105eb674e1151358921b7abaa",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then the hash check matches, no request is made and None is returned
        assert new_state is None

    def test_update_move_workspace(self, respx_mock, workspace2):
        respx_mock.delete("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(200, json={})
        )
        respx_mock.post("/api/api/workspaces/personal/workspace2/items").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the desired resource changes from workspace1 to workspace2
        # but the item has not changed
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace2,
            type="type1",
            description="a test item",
            name="item1",
            group="group1",
            format=3,
            content={"key1": "value1", "key2": True, "key3": 42},
        )
        old_state = SimpleNamespace(
            visibility="shared",
            name="item1",
            group="group1",
            workspace_name="workspace1",
            content_hash="21368e5f74f8fa8651d55f88d48ba651e3e8f0c105eb674e1151358921b7abaa",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then a new one is created
        request1 = respx_mock.calls[0].request
        assert json.loads(request1.content) == {
            "name": "item1",
            "description": "a test item",
            "format": 3,
            "group": "group1",
            "content": {"key1": "value1", "key2": True, "key3": 42},
            "type": "type1",
        }
        # and the existing item is deleted
        request2 = respx_mock.calls[1].request
        assert request2.method == "DELETE"
        assert request2.url.path == "/api/api/workspaces/shared/workspace1/items/group1/item1"
        # and the new state is returned with the same hash
        assert new_state == {
            "visibility": "personal",
            "name": "item1",
            "workspace_name": "workspace2",
            "group": "group1",
            "content_hash": "21368e5f74f8fa8651d55f88d48ba651e3e8f0c105eb674e1151358921b7abaa",
        }

    def test_update_move_workspace_failure(self, respx_mock, workspace2):
        # given we need to move the item to another workspace
        # but creation of the new one fails
        respx_mock.post("/api/api/workspaces/personal/workspace2/items").mock(
            return_value=httpx.Response(400, json={})
        )
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace2,
            type="type1",
            description="a test item",
            name="item1",
            group="group1",
            format=3,
            content={"key1": "value1", "key2": True, "key3": 42},
        )
        old_state = SimpleNamespace(
            visibility="shared",
            name="item1",
            group="group1",
            workspace_name="workspace1",
            content_hash="21368e5f74f8fa8651d55f88d48ba651e3e8f0c105eb674e1151358921b7abaa",
        )
        # when we call update it raises and no delete call is made
        with pytest.raises(httpx.HTTPError):
            sut.update(self.client, old_state)

    def test_update_rename_item(self, respx_mock, workspace1):
        respx_mock.delete("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(200, json={})
        )
        respx_mock.post("/api/api/workspaces/shared/workspace1/items").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the desired resource changes it's name
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            description="a test item",
            name="item_renamed",
            group="group1",
            format=3,
            content={"key1": "value1", "key2": True, "key3": 42},
        )
        old_state = SimpleNamespace(
            visibility="shared",
            name="item1",
            group="group1",
            workspace_name="workspace1",
            content_hash="21368e5f74f8fa8651d55f88d48ba651e3e8f0c105eb674e1151358921b7abaa",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then a new one is created
        request1 = respx_mock.calls[0].request
        assert json.loads(request1.content) == {
            "name": "item_renamed",
            "description": "a test item",
            "format": 3,
            "group": "group1",
            "content": {"key1": "value1", "key2": True, "key3": 42},
            "type": "type1",
        }
        # and the existing item is deleted
        request2 = respx_mock.calls[1].request
        assert request2.method == "DELETE"
        assert request2.url.path == "/api/api/workspaces/shared/workspace1/items/group1/item1"
        # and the new state is returned with a new hash
        assert new_state == {
            "visibility": "shared",
            "name": "item_renamed",
            "workspace_name": "workspace1",
            "group": "group1",
            "content_hash": "715e9305b154ce58b460f848ac0e65f7fc25657e0375c28b7a7843381c1df638",
        }

    def test_update_change_group(self, respx_mock, workspace1):
        respx_mock.delete("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(200, json={})
        )
        respx_mock.post("/api/api/workspaces/shared/workspace1/items").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the desired resource is in a different group
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            description="a test item",
            name="item_renamed",
            group="group2",
            format=3,
            content={"key1": "value1", "key2": True, "key3": 42},
        )
        old_state = SimpleNamespace(
            visibility="shared",
            name="item1",
            group="group1",
            workspace_name="workspace1",
            content_hash="21368e5f74f8fa8651d55f88d48ba651e3e8f0c105eb674e1151358921b7abaa",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then a new one is created
        request1 = respx_mock.calls[0].request
        assert json.loads(request1.content) == {
            "name": "item_renamed",
            "description": "a test item",
            "format": 3,
            "group": "group2",
            "content": {"key1": "value1", "key2": True, "key3": 42},
            "type": "type1",
        }
        # and the existing item is deleted
        request2 = respx_mock.calls[1].request
        assert request2.method == "DELETE"
        assert request2.url.path == "/api/api/workspaces/shared/workspace1/items/group1/item1"
        # and the new state is returned with a new hash
        assert new_state == {
            "visibility": "shared",
            "name": "item_renamed",
            "workspace_name": "workspace1",
            "group": "group2",
            "content_hash": "e348d17261f9b7d35c52e15f3b40d7aa220b0fd23d56415651721c269c088fd5",
        }

    def test_update_change_content(self, respx_mock, workspace1):
        respx_mock.put("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the desired resource changes the content of the item
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            description="a test item",
            name="item1",
            group="group1",
            format=3,
            content={"key1": "value_changed", "key2": True, "key3": 42},
        )
        old_state = SimpleNamespace(
            visibility="shared",
            name="item1",
            group="group1",
            workspace_name="workspace1",
            content_hash="21368e5f74f8fa8651d55f88d48ba651e3e8f0c105eb674e1151358921b7abaa",
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then the existing item is updated
        request1 = respx_mock.calls[0].request
        assert json.loads(request1.content) == {
            "name": "item1",
            "description": "a test item",
            "format": 3,
            "group": "group1",
            "content": {"key1": "value_changed", "key2": True, "key3": 42},
            "type": "type1",
        }
        # and the new state is returned with a new hash
        assert new_state == {
            "visibility": "shared",
            "name": "item1",
            "workspace_name": "workspace1",
            "group": "group1",
            "content_hash": "39a57eb19bae4c013829d9cc5be72d9796a44edbd2b15d4044b691321ee6e5a7",
        }

    def test_update_no_hash_on_existing(self, respx_mock, workspace1):
        respx_mock.put("/api/api/workspaces/shared/workspace1/items/group1/item1").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the desired resource is the same as the remote
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            description="a test item",
            name="item1",
            group="group1",
            format=3,
            content={"key1": "value1", "key2": True, "key3": 42},
        )
        # but the remote is from an old version without hashing
        old_state = SimpleNamespace(
            visibility="shared",
            name="item1",
            group="group1",
            workspace_name="workspace1"
        )
        # when we call update
        new_state = sut.update(self.client, old_state)
        # then the existing item is updated
        request1 = respx_mock.calls[0].request
        assert json.loads(request1.content) == {
            "name": "item1",
            "description": "a test item",
            "format": 3,
            "group": "group1",
            "content": {"key1": "value1", "key2": True, "key3": 42},
            "type": "type1",
        }
        # and the new state is returned with a new hash
        assert new_state == {
            "visibility": "shared",
            "name": "item1",
            "workspace_name": "workspace1",
            "group": "group1",
            "content_hash": "21368e5f74f8fa8651d55f88d48ba651e3e8f0c105eb674e1151358921b7abaa",
        }

    def test_dump_item(self, workspace1):
        # given an item resource referencing a workspace resource
        sut = workspace.WorkspaceItemResource(
            id="item1",
            workspace=workspace1,
            type="type1",
            description="a test item",
            name="item1",
            group="group1",
            format=3,
            content={"key1": "value1", "key2": True, "key3": 42},
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then the dumped has the workspace as a $ref
        assert dumped == {
            "name": "item1",
            "workspace": {"$ref": "wk1"},
            "type": "type1",
            "description": "a test item",
            "group": "group1",
            "format": 3,
            "content": {"key1": "value1", "key2": True, "key3": 42},
        }

    def test_undump_item(self):
        workspace_ref = workspace.WorkspaceRef(
            id="wrkref",
            visibility=workspace.Visibility.SHARED,
            name="workspace5"
        )
        # given the dumped form of an item
        dumped = {
            "name": "item1",
            "workspace": {"$ref": "wrkref"},
            "type": "type1",
            "description": "a test item",
            "group": "group1",
            "format": 3,
            "content": {"key1": "value1", "key2": True, "key3": 42},
        }
        # when we deserialize it
        result = workspace.WorkspaceItemResource.model_validate(
            dumped,
            context={
                "style": "dump",
                "$refs": {
                    "wrkref": workspace_ref
                },
                "id": "item_id"
            }
        )
        # then the workspace is connected
        assert result.workspace == workspace_ref
        # and the resource id gets set
        assert result.id == "item_id"
